//
//	capture.c
//
//	This is a DLL which is designed to be injected
//  into the address space of the process which owns the
//  window that needs to be "captured"
//
//	Once loaded, this DLL hooks BeginPaint and EndPaint.
//
//	The controlling program then invokes the "CaptureWindow" export, which
//  causes the application's window to repaint itself into our
//  specially constructed memory-device-context (instead of the normal
//  screen-device-context).
//
//	Don't even bother trying to test this with Internet Explorer,
//  for some reason it never draws into it's windows - instead it 
//  validates it's main window, then BitBlts directly to the screen
//  instead of it's device context for the window :-0
//
//	v1.0 - Mr Bigglesworth ;-)
//
#include <windows.h>
#include "memdc.h"

DWORD BeginPaintCallIdx;
DWORD EndPaintCallIdx;
DWORD pfnChainBeginPaint;
DWORD pfnChainEndPaint;

#pragma comment(linker, "/EXPORT:CaptureWindow=_CaptureWindow@4")
#pragma comment(linker, "/OPT:NOWIN98")

struct PAINTHOOK
{
	HWND	hWnd;			// the "base" window we want to capture
	RECT	base;			// the "base" outer rectangle of window
	HDC		hdcScreen;

	HDC		hdcMem;			// memory device-context
	HBITMAP	hBitmap;
	
	BOOL	fInvalidate;	// Force a repaint using InvalidateRect

} PaintHook;

//
//	Replacement for process-wide BeginPaint API (see MSDN)
//
HDC WINAPI BeginPaintHook(HWND hwnd, PAINTSTRUCT *ps)
{
	// are we painting the window we want to capture?
	if(hwnd == PaintHook.hWnd)
	{
		// BeginPaint needs to return a device-context into
		// which the application draws it's window.
		// Just return our "memory" device-context instead
		// When the application eventually calls EndPaint, we can see
		// what they painted and capture the window at that stage
		ZeroMemory(ps, sizeof(PAINTSTRUCT));

		OutputDebugString("Painting\n");

		ps->hdc = PaintHook.hdcMem;
		GetClientRect(hwnd, &ps->rcPaint);
		return ps->hdc;
	}
	else
	{
		__asm	mov		eax, BeginPaintCallIdx
		__asm	push	ps
		__asm	push	hwnd
		__asm	call	pfnChainBeginPaint
	}
}

//
//	Replacement for process-wide BeginPaint API (see MSDN)
//
BOOL WINAPI EndPaintHook(HWND hwnd, PAINTSTRUCT *ps)
{
	// are we painting the window we want to capture?
	if(hwnd == PaintHook.hWnd)
	{
		// The application has finished drawing into ps::hdc, 
		// (the same as our PaintHook.hdcMem), so we can use this
		// device-context at some later time to see what happened
		return TRUE;
	}
	else
	{
		__asm	mov		eax, EndPaintCallIdx
		__asm	push	ps
		__asm	push	hwnd
		__asm	call	pfnChainEndPaint
	}
}

//
//	Very specialized routine which hooks certain user32.dll APIs which 
//  have the form:
//
//		mov  eax, sys-call-index
//		mov  edx, parameter
//		call win32k-system-service-dispatcher
//
PBYTE HookSysCall(LPCTSTR szModule, LPCSTR szProc, PVOID pfnHook, DWORD *syscall)
{
	HMODULE hModule;
	PBYTE	pfnProc;

	hModule = GetModuleHandle(szModule);
	pfnProc = (PBYTE)GetProcAddress(hModule, szProc);

	if(pfnProc == 0)
		return NULL;

	// check that the first instruction is a "mov eax, imm32"
	if(pfnProc[0] == 0xb8)
	{
		DWORD oldprot;

		VirtualProtect(pfnProc, 5, PAGE_EXECUTE_READWRITE, &oldprot);

		// extract the system call index from the "mov eax" instruction
		*syscall = *(DWORD *)(pfnProc + 1);

		// replace the "mov eax,imm32" with a "jmp rel32"
		pfnProc[0]			    = 0xe9;
		*(DWORD *)(pfnProc + 1) = (DWORD)pfnHook - (DWORD)(pfnProc + 5);

		// return pointer to next instruction
		return (pfnProc+5);
	}
	else
	{
		return NULL;
	}
}

//
//
//
BOOL UnhookSysCall(LPCTSTR szModule, LPCSTR szProc, DWORD syscall)
{
	HMODULE hModule;
	PBYTE	pfnProc;

	hModule = GetModuleHandle(szModule);
	pfnProc = (PBYTE)GetProcAddress(hModule, szProc);

	if(pfnProc == 0)
		return FALSE;

	// is it one we hooked?
	if(pfnProc[0] == 0xe9)
	{
		DWORD oldprot;

		VirtualProtect(pfnProc, 5, PAGE_EXECUTE_READWRITE, &oldprot);

		// reconstruct the normal syscall
		pfnProc[0]				= 0xb8;		// mov eax, imm32
		*(DWORD *)(pfnProc + 1) = syscall;

		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

int RectWidth(RECT *rect)
{
	return rect->right - rect->left;
}

int RectHeight(RECT *rect)
{
	return rect->bottom - rect->top;
}

BOOL _PrintWindow(HWND hwnd, HDC hdc, UINT nFlags)
{
	typedef BOOL (WINAPI * PWP)(HWND, HDC, UINT);
	static PWP pfnPrintWindow;
	HMODULE hWinSpool;
	BOOL fTried = FALSE;

	if(fTried == FALSE)
	{
		hWinSpool = LoadLibrary("winspool.dll");

		pfnPrintWindow = (PWP)GetProcAddress(hWinSpool, "PrintWindow");
		fTried = TRUE;
	}

	if(pfnPrintWindow)
	{
		return pfnPrintWindow(hwnd, hdc, nFlags);
	}
	else
	{
		return FALSE;
	}
}

//
//	Capture the specified window contents.
//
//	hwnd - window to capture
//  hdc  - device-context to store the capture
//  base - coordinates of the main window we are capturing
//			(and indirectly the size of the device-context, above)
//
static BOOL CaptureWindowToMemDC(HWND hwnd, HDC hdc, RECT *base)
{
	RECT	rect;
	RECT	client;
	int		state;
	MEMDC	memdc;

	//
	//	NONCLIENT AREA
	//
	//	Cause the nonclient area to be painted into our device-context
	//
	GetWindowRect(hwnd, &rect);

	memdc.init(PaintHook.hdcScreen, RectWidth(&rect), RectHeight(&rect));

	if(!_PrintWindow(hwnd, memdc, 0))
		SendMessage(hwnd, WM_PRINT, memdc, PRF_NONCLIENT);

	BitBlt(hdc, (rect.left-base->left),
				(rect.top-base->top), 
				RectWidth(&rect), 
				RectHeight(&rect), 
				memdc, 0, 0, SRCCOPY);

	//
	//	CLIENT AREA
	//
	GetClientRect(hwnd, &client);
	ClientToScreen(hwnd, (POINT *)&client.left);
	ClientToScreen(hwnd, (POINT *)&client.right);

	memdc.init(PaintHook.hdcScreen, RectWidth(&client), RectHeight(&client));

	PaintHook.hdcMem = memdc;	
	PaintHook.hWnd	 = hwnd;

	//
	// First erase the client-area background
	//
	if(!_PrintWindow(hwnd, memdc, 1))
	{
	state = SaveDC(memdc);
	SendMessage(hwnd, WM_ERASEBKGND, memdc, 0);
	RestoreDC(memdc, state);

	//
	// Now paint the window content - this is where the BeginPaint/EndPaint get called
	//
	if(PaintHook.fInvalidate)
	{
		InvalidateRect(hwnd, 0, FALSE);
		UpdateWindow(hwnd);
	}
	else
	{
		SendMessage(hwnd, WM_PAINT, 0, 0);
	}
	}

	BitBlt(hdc, (client.left-base->left),
				(client.top-base->top),
				RectWidth(&client), 
				RectHeight(&client),
				memdc, 0, 0, SRCCOPY);

	PaintHook.hWnd	 = 0;
	return FALSE;
} 

BOOL DoHooking()
{
	pfnChainBeginPaint = (DWORD)HookSysCall("user32.dll", "BeginPaint", BeginPaintHook, &BeginPaintCallIdx);
	pfnChainEndPaint   = (DWORD)HookSysCall("user32.dll", "EndPaint",   EndPaintHook,   &EndPaintCallIdx);

	return pfnChainBeginPaint && pfnChainEndPaint;
}

void UndoHooking()
{
	UnhookSysCall("user32.dll", "BeginPaint", BeginPaintCallIdx);
	UnhookSysCall("user32.dll", "EndPaint",   EndPaintCallIdx);
}

BOOL CALLBACK CaptureChildren(HWND hwnd, HDC hdc)
{
	DWORD dwStyle   = GetWindowLong(hwnd, GWL_STYLE);
	DWORD dwStyleEx = GetWindowLong(hwnd, GWL_EXSTYLE);

	if((dwStyle & WS_VISIBLE) == 0)
		return TRUE;

	if(dwStyleEx & WS_EX_TRANSPARENT)
		return TRUE;

	CaptureWindowToMemDC(hwnd, hdc, &PaintHook.base);
	return TRUE;
}

//
//	CaptureWindow - Main API of this DLL
//
//	hwnd - handle to window to capture
//
extern "C"
DWORD WINAPI CaptureWindow(HWND hwnd)
{
	RECT rect;
	WINDOWPLACEMENT wp = { sizeof(wp) };

	MEMDC	memdc;
	
	PaintHook.hdcScreen = GetDC(0);

	// find out how big the window is..even if it is minimized
	//GetWindowRect(hwnd, &rect);
	GetWindowPlacement(hwnd, &wp);
	rect = wp.rcNormalPosition;

	PaintHook.base			= rect;
	PaintHook.fInvalidate	= TRUE;

	//
	//	Create a main "hidden" memory-DC which we will use to
	//  paint each window and child window into
	//
	memdc.init(PaintHook.hdcScreen, RectWidth(&rect), RectHeight(&rect));

	// Capture the window we are interested in
	CaptureWindowToMemDC(hwnd, memdc, &PaintHook.base);

	// Also need to capture ALL of it's child windows
	EnumChildWindows(hwnd, (WNDENUMPROC)CaptureChildren, (LPARAM)(HDC)memdc);

	// To see what was painted, copy the bitmap to the Windows clipboard!!
	// a real capture DLL should create an on-disk bitmap or something
	// like that....i.e. make a function SaveBitmap(HBITMAP hBitmap, char *file)
	if(OpenClipboard(NULL))
	{
		EmptyClipboard();
		SetClipboardData(CF_BITMAP, memdc.bitmap());
		CloseClipboard();
	}

	// Cleanup
	ReleaseDC(0, PaintHook.hdcScreen);
	return TRUE;
}

BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved)
{
	switch(dwReason)
	{
	case DLL_PROCESS_ATTACH:
		return DoHooking();

	case DLL_PROCESS_DETACH:
		UndoHooking();
		break;
	}

	return TRUE;
}